# """
# Assign XTB calculator to ase_atoms object following defualts or user input!
# Written by Michael Taylor

# This module is largely depricated in architector in favor of io_calc.py
# It is useful still for xtb evalulations away from architector's main codebase.
# """

import numpy as np

from xtb.ase.calculator import XTB

from architector import arch_context_manage, io_molecule
import architector.io_ptable as io_ptable

import ase
from ase.optimize import BFGSLineSearch
from ase.atoms import Atom,Atoms

def set_XTB_calc(ase_atoms, parameters=dict(), assembly=False, isCp_lig=False):
    """set_XTB_calc 
    assign xtb calculator to atoms instance!

    Parameters
    ----------
    ase_atoms : ase.atoms.Atoms
        atoms to assign calculator to
    parameters : dict, optional
        parameters from input/ io_process_input!
    assembly : bool, optional
        whether or not this is assembly or final relaxation, by default False
    isCP_lig : bool, optional
        whether this is cP ligand evaluation or not, by default False
    """
    if isCp_lig:
        ase_atoms.set_initial_charges(np.zeros(len(ase_atoms)))
        ase_atoms.set_initial_magnetic_moments(np.zeros(len(ase_atoms)))
        calc = XTB(method="GFN-FF") # Defaul to only GFN-FF for ligand conformer relaxation.
    else:
        # Charge -> charges already assigned to components during assembly
        if (parameters['full_charge'] is not None) and (not assembly):
            charge_vect = np.zeros(len(ase_atoms))
            charge_vect[0] = parameters['full_charge']
            ase_atoms.set_initial_charges(charge_vect)
        else:
            charge_vect = ase_atoms.get_initial_charges()

        mol_charge = np.sum(charge_vect)
        symbols = ase_atoms.get_chemical_symbols()
        metals = [x for x in symbols if x in io_ptable.all_metals]

        f_in_core = False
        
        if len(metals) == 1:
            if metals[0] in io_ptable.heavy_metals:
                f_in_core = True
        else:
            print('No metals - continuing anyway with obviously no f in core elements.')

        # Handle spin / magnetism
        even_odd_electrons = (np.sum([atom.number for atom in ase_atoms])-mol_charge) % 2
        if (parameters['full_spin'] is not None) and (not assembly):
            uhf = parameters['full_spin']
            uhf_start = np.zeros(len(ase_atoms))
            uhf_start[0] = uhf
            ase_atoms.set_initial_magnetic_moments(uhf_start)
        else:
            uhf = parameters['metal_spin'] # Metal spin set by io_process_input to defaults.
            if (even_odd_electrons == 1) and (uhf == 0):
                uhf = 1
            elif (even_odd_electrons == 1) and (uhf < 7) and (uhf % 2 == 0):
                uhf += 1
            elif (even_odd_electrons == 1) and (uhf >= 7) and (uhf % 2 == 0):
                uhf -= 1
            if (even_odd_electrons == 0) and (uhf % 2 == 1):
                uhf = uhf - 1 
            elif (even_odd_electrons == 1) and (uhf % 2 == 0):
                uhf = uhf + 1
            uhf_start = np.zeros(len(ase_atoms))
            if not f_in_core:
                uhf_start[0] = uhf
            else: # F in core assumes for a 3+ lanthanide there are 11 valence electrons (8 once the 3+ is taken into account)
                even_odd_electrons = (np.sum([atom.number for atom in ase_atoms]))
                even_odd_electrons = even_odd_electrons - io_ptable.elements.index(metals[0]) + 11 - mol_charge
                even_odd_electrons = even_odd_electrons % 2
                if (even_odd_electrons == 0):
                    uhf_start[0] = 0
                else:
                    uhf_start[0] = 1
            ase_atoms.set_initial_magnetic_moments(uhf_start)

        if assembly:
            if parameters['assemble_method'] == 'GFN-FF': # Need to turn off charges for GFN-FF evaluation. Probably an XTB-end bug.
                ase_atoms.set_initial_charges(np.zeros(len(ase_atoms)))
                ase_atoms.set_initial_magnetic_moments(np.zeros(len(ase_atoms)))
            calc = XTB(method=parameters['assemble_method'], solvent=parameters['solvent'])
        else:
            if parameters['full_method'] == 'GFN-FF': # Need to turn off charges for GFN-FF evaluation. Probably an XTB-end bug.
                ase_atoms.set_initial_charges(np.zeros(len(ase_atoms)))
                ase_atoms.set_initial_magnetic_moments(np.zeros(len(ase_atoms)))
            calc = XTB(method=parameters['full_method'], solvent=parameters['solvent'])

    #########################################################
    ########### Calculator Now Set! #########################
    #########################################################
    ase_atoms.calc = calc
    return ase_atoms

def set_XTB_calc_lig(ase_atoms, charge=None, uhf=None, method='GFN2-xTB',solvent='none'):
    """set_XTB_calc 
    assign xtb calculator to ligand atoms instance!

    Parameters
    ----------
    ase_atoms : ase.atoms.Atoms
        atoms to assign calculator to
    charge : int, optional
        charge of the species, default to initial charges set on ase_atoms
    uhf : int, optional
        number of unpaired electrons in the system, default to 0
    method : str, optional
        which gfn family method to use, default GFN2-xTB
    solvent: str, optional
        use a solvent?, default 'none'
    """
    if charge:
        mol_charge = charge
    else:
        mol_charge = np.sum(ase_atoms.get_initial_charges())

    charge_vect = np.zeros(len(ase_atoms))
    charge_vect[0] = mol_charge
    ase_atoms.set_initial_charges(charge_vect)
        
    # Handle spin / magnetism
    even_odd_electrons = (np.sum([atom.number for atom in ase_atoms])-mol_charge) % 2
    if (uhf is not None):
        uhf = uhf
        if (even_odd_electrons == 1) and (uhf == 0):
            uhf = 1
        elif (even_odd_electrons == 1) and (uhf < 7) and (uhf % 2 == 0):
            uhf += 1
        elif (even_odd_electrons == 1) and (uhf >= 7) and (uhf % 2 == 0):
            uhf -= 1
        if (even_odd_electrons == 0) and (uhf % 2 == 1):
            uhf = uhf - 1 
        elif (even_odd_electrons == 1) and (uhf % 2 == 0):
            uhf = uhf + 1
        uhf_start = np.zeros(len(ase_atoms))
        uhf_start[0] = uhf
        ase_atoms.set_initial_magnetic_moments(uhf_start)
    else:
        uhf = 0 # Set spin to LS by default
        if (even_odd_electrons == 1) and (uhf == 0):
            uhf = 1
        elif (even_odd_electrons == 1) and (uhf < 7) and (uhf % 2 == 0):
            uhf += 1
        elif (even_odd_electrons == 1) and (uhf >= 7) and (uhf % 2 == 0):
            uhf -= 1
        if (even_odd_electrons == 0) and (uhf % 2 == 1):
            uhf = uhf - 1 
        elif (even_odd_electrons == 1) and (uhf % 2 == 0):
            uhf = uhf + 1
        uhf_start = np.zeros(len(ase_atoms))
        uhf_start[0] = uhf
        ase_atoms.set_initial_magnetic_moments(uhf_start)

    if method == 'GFN-FF': # Need to turn off charges for GFN-FF evaluation. Probably an XTB-end bug.
        ase_atoms.set_initial_charges(np.zeros(len(ase_atoms)))
        ase_atoms.set_initial_magnetic_moments(np.zeros(len(ase_atoms)))
    calc = XTB(method=method, solvent=solvent)

    #########################################################
    ########### Calculator Now Set! #########################
    #########################################################
    ase_atoms.calc = calc

def set_XTB_calc_straight(ase_atoms, charge=None, uhf=None, method='GFN2-xTB',solvent='none'):
    """set_XTB_calc_straight
    assign xtb calculator atoms object with exaclty spin/charge requeste

    Parameters
    ----------
    ase_atoms : ase.atoms.Atoms
        atoms to assign calculator to
    charge : int, optional
        charge of the species, default to initial charges set on ase_atoms
    uhf : int, optional
        number of unpaired electrons in the system, default to 0
    method : str, optional
        which gfn family method to use, default GFN2-xTB
    solvent: str, optional
        use a solvent?, default 'none'
    """
    if charge:
        mol_charge = charge
    else:
        mol_charge = np.sum(ase_atoms.get_initial_charges())

    charge_vect = np.zeros(len(ase_atoms))
    charge_vect[0] = mol_charge
    ase_atoms.set_initial_charges(charge_vect)
        
    uhf_start = np.zeros(len(ase_atoms))
    uhf_start[0] = uhf
    ase_atoms.set_initial_magnetic_moments(uhf_start)

    if method == 'GFN-FF': # Need to turn off charges for GFN-FF evaluation. Probably an XTB-end bug.
        ase_atoms.set_initial_charges(np.zeros(len(ase_atoms)))
        ase_atoms.set_initial_magnetic_moments(np.zeros(len(ase_atoms)))
    calc = XTB(method=method, solvent=solvent)

    #########################################################
    ########### Calculator Now Set! #########################
    #########################################################
    ase_atoms.calc = calc


def xtb_relax(structure, charge=None, uhf=None, method='GFN2-xTB',solvent='none',fmax=0.05,
            detect_charge_spin=False):
    """xtb_relax relax the structure with xtb

    Parameters
    ----------
    structure : any 3D structure
       xyz, mol2string, ase atoms ...
    charge :int, optional
        total charge on the system, by default None
    uhf : int, optional
        number of unpaired electrons in the system, by default None
    method : str, optional
        which method to use, by default 'GFN2-xTB'
    solvent : str, optional
        any name xtb solvent, by default 'none'
    fmax : float, optional
        default 0.05 eV/Angstrom
    detect_charge_spin : bool, optional
        Use obmol and io_ptable metal defaults to assign charges and spins?, default False.

    Returns
    -------
    ase_atoms : ase.atoms.Atoms
        relaxed structure
    good : bool
        whether the relaxation was succesful!
    """
    if isinstance(structure,ase.atoms.Atoms):
        ase_atoms = structure
        if ase_atoms.calc is None:
            set_XTB_calc_straight(ase_atoms,charge=charge,uhf=uhf,method=method,solvent=solvent) 
    else:
        mol = io_molecule.convert_io_molecule(structure,detect_charge_spin=detect_charge_spin)
        ase_atoms = mol.ase_atoms
        if detect_charge_spin:
            set_XTB_calc_straight(ase_atoms,charge=mol.charge,uhf=mol.xtb_uhf,method=method,solvent=solvent) 
        else:
            set_XTB_calc_straight(ase_atoms,charge=charge,uhf=uhf,method=method,solvent=solvent) 
    good = True
    with arch_context_manage.make_temp_directory() as _:
        try:
            dyn = BFGSLineSearch(ase_atoms)
            dyn.run(fmax=fmax)
        except:
            good = False
    return ase_atoms,good


def xtb_sp(structure, charge=None, uhf=None, method='GFN2-xTB',solvent='none',
            detect_charge_spin=False):
    """xtb_sp singlepoint on the structure with xtb

    Parameters
    ----------
    structure : any 3D structure
       xyz, mol2string, ase atoms ...
    charge :int, optional
        total charge on the system, by default None
    uhf : int, optional
        number of unpaired electrons in the system, by default None
    method : str, optional
        which method to use, by default 'GFN2-xTB'
    solvent : str, optional
        any name xtb solvent, by default 'none'

    Returns
    -------
    ase_atoms : ase.atoms.Atoms
        structure with calculator/energy calculated
    good : bool
        whether the relaxation was succesful!
    """
    if isinstance(structure,ase.atoms.Atoms):
        ase_atoms = structure
        if ase_atoms.calc is None:
            set_XTB_calc_straight(ase_atoms,charge=charge,uhf=uhf,method=method,solvent=solvent) 
    else:
        mol = io_molecule.convert_io_molecule(structure,detect_charge_spin=detect_charge_spin)
        ase_atoms = mol.ase_atoms
        if detect_charge_spin:
            set_XTB_calc_straight(ase_atoms,charge=mol.charge,uhf=mol.xtb_uhf,method=method,solvent=solvent) 
        else:
            set_XTB_calc_straight(ase_atoms,charge=charge,uhf=uhf,method=method,solvent=solvent) 
    good = True
    with arch_context_manage.make_temp_directory() as _:
        try:
            ase_atoms.get_total_energy()
        except:
            good = False
    return ase_atoms,good


eV2Hartree = 1 / 27.2114

def get_rxyz_string(ase_atoms):
    """dump the ase_atoms to rxyz file string"""
    natom = len(ase_atoms)
    ss = ''
    # write an xyz file first
    ss += "%d\n\n" % natom
    for symb, coord in zip(ase_atoms.get_chemical_symbols(), ase_atoms.positions):
        ss += "%2s %12.6f %12.6f %12.6f\n" % (symb, *coord[:3])
    ss += "\n"
    # then force
    ss += "FORCES\n"
    for symb, force in zip(ase_atoms.get_chemical_symbols(), ase_atoms.get_forces()):
        ss += "%3s %22.14e %22.14e %22.14e\n" % (symb, *(force*eV2Hartree))
    ss += '\n'
    # then pbc, if needed
    # Potentially usefull - ase_atoms.cell, ase_atoms.get_pbc(), ase_atoms.get_cell_lengths_and_angles(), get_celldisp(), get_cell()
    # if ase_atoms.cell is not None:
    #     ss += "PBC\n"
    #     ss += "%12.6f %12.6f %12.6f\n" % (chemical.pbc.boxhi[0] -
    #         chemical.pbc.boxlo[0], 0.0, 0.0))
    #     ss += "%12.6f %12.6f %12.6f\n" % (chemical.pbc.xy,
    #         chemical.pbc.boxhi[1] - chemical.pbc.boxlo[1], 0.0))
    #     fp.write("%12.6f %12.6f %12.6f\n" % (chemical.pbc.xz, 
    #                                             chemical.pbc.yz, 
    #                     chemical.pbc.boxhi[2] - chemical.pbc.boxlo[2]))
    #     fp.write('\n')
    # then charge and spin
    ss += "ATOM-CHARGE-SPIN\n"
    for symb, charge, spin in zip(ase_atoms.get_chemical_symbols(), ase_atoms.get_charges(),
                                    ase_atoms.get_initial_magnetic_moments()):
        ss += "%2s %12.6f %12.6f\n" % (symb, charge, spin)
    ss += '\n'
    # then different properties
    ss += "ENERGY %22.14f\n" % (ase_atoms.get_total_energy()*eV2Hartree)
    ss += "PA_BINDING_ENERGY %22.14f\n" % (ase_atoms.get_total_energy()*eV2Hartree
                                            / len(ase_atoms))
    dipole_vect = ase_atoms.get_dipole_moment()
    ss += 'DIPOLE_VEC %12.6f %12.6f %12.6f\n' % (dipole_vect[0],
                                        dipole_vect[1],
                                        dipole_vect[2])
    ss += 'DIPOLE %12.6f\n'%np.linalg.norm(dipole_vect)
    ss += "CHARGE %d\n" % ase_atoms.get_initial_charges().sum()
    ss += "MULTIPLICITY %d\n" % round(sum(ase_atoms.get_initial_magnetic_moments)+1)
    # fp.write("DIFFAB %f\n" % sum(result['atomspin']))
    # fp.write('BAND_GAP %12.6f\n' % (result["egap"] * eV2Hartree))
    # fp.write('CONVERGED %s\n' % str(result["qconverged"]))
    # for prop in more_props:
    #     fp.write('%s %f\n' % (prop.upper(), result[prop]))
    ss += 'SOURCE %s\n' % ase_atoms.get_calculator().name
    return ss


def calc_xtb_ref_dict():
    energydict = dict()
    print('--------1---------')
    for i,elem in enumerate(io_ptable.elements):
        if i>0 and i<87:
            atoms = Atoms([Atom(elem,(0,0,0))])
            if elem in io_ptable.all_metals:
                if elem in io_ptable.lanthanides:
                    spin = 0
                    charge = 3
                else:
                    spin = io_ptable.metal_spin_dict[elem]
                    charge = io_ptable.metal_charge_dict[elem]
            else:
                if i % 2 == 0:
                    spin = 0
                else:
                    spin = 1 
                charge = 0
            set_XTB_calc_straight(atoms,charge=charge,uhf=spin)
            try:
                energy = atoms.get_total_energy()
                energydict[elem] = energy
            except:
                energydict[elem] = None
    print('--------2---------')
    for elem,spin in io_ptable.second_choice_metal_spin_dict.items():
        print(elem)
        if io_ptable.elements.index(elem) < 87:
            atoms = Atoms([Atom(elem,(0,0,0))])
            charge = io_ptable.metal_charge_dict[elem]
            set_XTB_calc_straight(atoms,charge=charge,uhf=spin)
            try: 
                energy = atoms.get_total_energy()
            except:
                energy = None  
            failed=False
            if (energydict[elem] is None) and isinstance(energy,float):
                energydict[elem] = energy
                print('Fixed {} with alternate spin'.format(elem))
            elif (energydict[elem] is None) and (energy is None):
                failed = True
                print('{}'.format(elem) + ' STILL FAILED')
            if (not failed) and (energy is not None):
                if energydict[elem] > energy:
                    energydict[elem] = energy
                    print('replaced {} with alternate spin'.format(elem))
    print('--------3---------')
    for elem,spin in io_ptable.metal_spin_dict.items():
        print(elem)
        if io_ptable.elements.index(elem) < 87:
            atoms = Atoms([Atom(elem,(0,0,0))])
            charge = 0
            if (i % 2 == 0) or (elem not in io_ptable.lanthanides):
                spin =0
            else:
                spin =1
            set_XTB_calc_straight(atoms,charge=charge,uhf=spin)
            try: 
                energy = atoms.get_total_energy()
            except:
                energy = None  
            failed=False
            if (energydict[elem] is None) and isinstance(energy,float):
                energydict[elem] = energy
                print('Fixed {} with 0 charge low spin'.format(elem))
            elif (energydict[elem] is None) and (energy is None):
                failed = True
                print('{}'.format(elem) + ' STILL FAILED')
            if (not failed) and (energy is not None):
                if energydict[elem] > (energy):
                    energydict[elem] = energy
                    print('replaced {} with 0 charge alternate spin'.format(elem))
    print('--------4---------')
    for elem,spin in io_ptable.metal_spin_dict.items():
        print(elem)
        if io_ptable.elements.index(elem) < 87:
            atoms = Atoms([Atom(elem,(0,0,0))])
            charge = 0
            if (i % 2 == 0) or (elem not in io_ptable.lanthanides):
                spin =2
            else:
                spin =3
            set_XTB_calc_straight(atoms,charge=charge,uhf=spin)
            try: 
                energy = atoms.get_total_energy()
            except:
                energy = None  
            failed=False
            if (energydict[elem] is None) and isinstance(energy,float):
                energydict[elem] = energy
                print('Fixed {} with 0 charge 2 spin'.format(elem))
            elif (energydict[elem] is None) and (energy is None):
                failed = True
                print('{}'.format(elem) + ' STILL FAILED')
            if (not failed) and (energy is not None):
                if energydict[elem] > energy:
                    energydict[elem] = energy
                    print('replaced {} with 0 charge 2 spin'.format(elem))
    print('--------5---------')
    for elem,spin in io_ptable.metal_spin_dict.items():
        print(elem)
        if io_ptable.elements.index(elem) < 87:
            atoms = Atoms([Atom(elem,(0,0,0))])
            charge = 0
            if (i % 2 == 0) or (elem not in io_ptable.lanthanides):
                spin=4
            else:
                spin=5
            set_XTB_calc_straight(atoms,charge=charge,uhf=spin)
            try: 
                energy = atoms.get_total_energy()
            except:
                print('Failed 5 spin',elem)
                energy = None  
            failed=False
            if (energydict[elem] is None) and isinstance(energy,float):
                energydict[elem] = energy
                print('Fixed {} with 0 charge 4 spin'.format(elem))
            elif (energydict[elem] is None) and (energy is None):
                failed = True
                print('{}'.format(elem) + ' STILL FAILED')
            if (not failed) and (energy is not None):
                if energydict[elem] > energy:
                    energydict[elem] = energy
                    print('replaced {} with 0 charge 5 spin'.format(elem))

    # Fix Mo
    atoms = Atoms([Atom("Mo",(0,0,0))])
    charge=0
    set_XTB_calc_straight(atoms,charge=charge,uhf=6)
    e = atoms.get_total_energy()
    energydict['Mo'] = e
    # energydict_scaled_ha = {sym:val/Hartree for sym,val in energydict.items()}
    return energydict


""" 
Code for aligning molecules with arbitrary order. (Allowing permutations)
Applies by default to core of graoh depth 3 from the metal to align molecules.

Relies on a couple algorithms:
https://en.wikipedia.org/wiki/Hungarian_algorithm 
https://en.wikipedia.org/wiki/Root-mean-square_deviation_of_atomic_positions
https://en.wikipedia.org/wiki/Kabsch_algorithm 
A lot of shared ideas from:
https://pypi.org/project/rmsd/ 


Developed by Michael Taylor
"""
from multiprocessing.sharedctypes import Value
import numpy as np
import scipy
import copy

from scipy.optimize import linear_sum_assignment
from scipy.spatial.transform import Rotation as Rot
from architector import io_ptable

from architector.io_molecule import (convert_io_molecule, convert_ase_xyz)

def get_graph_depths(graph):
    """get_graph_depths 
    calculate all graph depths from molecular graph

    Returns
    -------
    depths : np.ndarray
        NXN matrix containing the graph depths of each atom to the others.
    """
    depths = scipy.sparse.csgraph.dijkstra(scipy.sparse.csgraph.csgraph_from_dense(graph))
    return depths

#### Permutation ordering and rmsd loss assignment adapted LANL code developed by Chang Liu

def permutation_cost_mat(pt_list1, pt_list2, label1, label2, costtype='xyz'):
    """permutation_cost_mat 
    return the cost matrix required by Hungarian method
    
    Only permutation with same label is allowed

    Parameters
    -------
    pt_list1 : np.ndarray
        array of coordinates of 1 
    pt_list2 : np.ndarray
        array of coordinates of 2 
    label1 : list/np.ndarray
        atom types or symols tied to coordinates 1
    label2 : list/np.ndarray
        atom types or symols tied to coordinates 2
    costtype : str, optional
        what type of cost/distance loss to use, by default 'xyz'
        'xyz' - pairwise distance from point of 1 to point in 2
        'COM' - distance to the center of mass

    Returns
    -------
    cost_mat : np.ndarray
        2D array NXN containing costs for the hungarian method
    """
    npt = len(pt_list1)
    cost_mat = np.zeros([npt, npt])
    pt_list1_com = np.mean(pt_list1,axis=0)
    pt_list2_com = np.mean(pt_list2,axis=0)
    for i, (xyz1, label1_t) in enumerate(zip(pt_list1, label1)):
        for j, (xyz2, label2_t) in enumerate(zip(pt_list2, label2)):
            if label1_t != label2_t:
                # permutation between different element is not allowed
                cost_mat[i,j] = np.inf
            else:
                if costtype == 'xyz':
                    diffvec = xyz1 - xyz2
                elif costtype == 'COM':
                    diffvec = (xyz1-pt_list1_com) - (xyz2-pt_list2_com)
                cost_mat[i,j] = np.dot(diffvec, diffvec)
    return cost_mat


def rmsd_align(tarmol, insrcmol, in_place=False):
    """rmsd_align 
    Perform alignment between two ase Atoms assuming indices are ideally permuted
    Return the rmsd_loss for the rotation.

    Parameters
    ----------
    tarmol : ase.atoms.Atoms
        reference/target molecule to match the generated molecule to.
    insrcmol : ase.atoms.Atoms
        generated/predicted molecule to match to reference/target molecule.
    in_place : bool, optional
        [description], by default False

    Returns
    -------
    rmsd_loss : float
        rmsd loss of the rotated version.
    r : scipy.spatial.transform.Rotation
        rotation instance to replicate matching the vectors.
    insrcmol : ase.atoms.Atoms
        rotated instance of the input molecule
    """
    insrcmol = insrcmol.copy()
    if not in_place:
        r = Rot.align_vectors(tarmol.positions, insrcmol.positions)
        newposits = r[0].apply(insrcmol.positions)
        insrcmol.set_positions(newposits)
    else: 
        r = None
    rmsd_loss = np.sqrt(np.sum((insrcmol.positions-tarmol.positions)**2)/len(tarmol))
    return rmsd_loss, r, insrcmol


def permute_align(tarmol, srcmol, maxiter=1, tol=1e-6, in_place=False):
    """permute_align 
    permute the atom order in mol to minimize the rmsd, in place

    A follow up rmsd alignment will be applied to the new molecule to
    calculate the new rmsd

    Parameters
    ----------
    tarmol : ase.atoms.Atoms
        reference/target molecule to match the generated molecule to.
    srcmol : ase.atoms.Atoms
        generated/predicted molecule to match to reference/target molecule.
    maxiter : int, optional
        maximum times permutation/rotation is repeated, by default 1
    tol : float, optional
        convergence tolerance, by default 1e-6
    in_place : bool, optional
        whether to perform permute with no rotation to calculate rsmd, by default False

    Returns
    -------
    rms : float
        rmsd of the permutation/rotation
    outr : scipy.spatial.transform.Rotation
        rotation instance required to align the molecules
    srcmol_1 : ase.atoms.Atoms
        rotated source molecule.
    """
    tarmol_1 = tarmol.copy()
    srcmol_1 = srcmol.copy()
    last_rms = np.inf
    outr = None
    count = 0
    for _ in range(maxiter):
        cost_mat = permutation_cost_mat(tarmol_1.positions, srcmol_1.positions, 
                                        tarmol_1.get_atomic_numbers(), srcmol_1.get_atomic_numbers(),
                                       costtype='xyz')
        permute = linear_sum_assignment(cost_mat)[1]
        srcmol_1 = srcmol_1[permute]
        rms, r, srcmol_1 = rmsd_align(tarmol_1, srcmol_1, in_place=in_place)
        if (count == 0) and (not in_place):
            outr = r[0] 
        elif (not in_place):
            outr = outr * r[0] # Composite rotation.
        else:
            outr = r
        if abs(rms-last_rms) < tol:
            break
        last_rms = rms
        count += 1
    return rms, outr, srcmol_1


def mirror_align(tarmol, srcmol, maxiter=1, tol=1e-6):
    """mirror_align 
    try mirror image alignment of molecule.

    Parameters
    ----------
    tarmol : ase.atoms.Atoms
        target molecule
    srcmol : ase.atoms.Atoms
        source molecule
    maxiter : int, optional
        how many times to try, by default 10
    tol : float, optional
        tolerance for convergence, by default 1e-2

    Returns
    -------
    rmsd : float
        rmsd value of converged mirror permutation
    outr : scipy.spatial.transform.Rotation
        rotation instance of best mirror image rotation.
    msrcmol : ase.atoms.Atoms
        mirrored rotated version of the molecule.
    """
    srcmol_tmp = srcmol.copy()
    newposits = srcmol_tmp.positions
    newposits[:,0] = -newposits[:,0]
    srcmol_tmp.set_positions(newposits)
    rmsd, outr, msrcmol = permute_align(tarmol,srcmol_tmp,maxiter=maxiter,tol=tol)
    return rmsd, outr, msrcmol
    

def calc_rmsd(genMol, compareMol, coresize=2, maxiter=1, sample=300, atom_types=None,
              return_structures=False, rmsd_type='simple'): 
    """calc_rmsd 
    Calculate the rmsd by different methods for this molecule compared to another.

    Parameters
    ----------
    genMol : str/Molecule
        molecule generated to rotate to reference.
    compareMol : str/Molecule
        molecule to compare to (reference molecule)
    rmsd_type : str, optional
        which type of rmsd to calculate , by default 'simple', possible 'edge'
    coresize : int, optional
        number of graph hops from the metal to consider when aligning molecules, by default 3 
    maxiter : int, optional
        number of iterations per time on permutation/alignment routine, by default 1
    sample : int, optional
        number of random rotations to sample initially to ensure best mapping between cores, by default 30
    atom_types: list, optional
        which atom types to consider for alignment - will default to full alignment, default None
    return_structures : bool, optional
        whether or not to return the rotated versions of the core and full structures, by defualt False

    Returns
    -------
    rmsd : float
        RMSD per atom comparison between the molecules.
    """

    genMol = convert_io_molecule(genMol)
    compareMol = convert_io_molecule(compareMol)

    # Check that these are stoichiometrically identical molecules.
    if np.any(sorted(genMol.ase_atoms.get_atomic_numbers()) != sorted(compareMol.ase_atoms.get_atomic_numbers())):
        print('Warning - comparison not possible between molecules of different sizes/stoichiometries.')
        flag_struct = True
        # Set ordering to be identical based on canonical labels.
        genMol_metalind = genMol.find_metal()
        genMol_graph_depths = get_graph_depths(genMol.graph)

        compareMol_metalind = compareMol.find_metal()
        compareMol_graph_depths = get_graph_depths(compareMol.graph)

        # Pull out center of molecule up to depth coresize graph hops for matching to reference.
        genMol_subset_component_inds = np.where(genMol_graph_depths[genMol_metalind] <= coresize)[0]
        compareMol_subset_component_inds = np.where(compareMol_graph_depths[compareMol_metalind] <= coresize)[0]

        tmp_self_comp = genMol.ase_atoms[genMol_subset_component_inds].copy() 
        tmp_ref_comp = compareMol.ase_atoms[compareMol_subset_component_inds].copy() 
        outcore = tmp_self_comp
        rmsd_loss_core = 1000
        rmsd_loss_full = 1000
    else:
        # Set ordering to be identical based on canonical labels.
        genMol_metalind = genMol.find_metal()
        genMol_graph_depths = get_graph_depths(genMol.graph)

        compareMol_metalind = compareMol.find_metal()
        compareMol_graph_depths = get_graph_depths(compareMol.graph)

        # Pull out center of molecule up to depth coresize graph hops for matching to reference.
        genMol_subset_component_inds = np.where(genMol_graph_depths[genMol_metalind] <= coresize)[0]
        compareMol_subset_component_inds = np.where(compareMol_graph_depths[compareMol_metalind] <= coresize)[0]

        tmp_self_comp = genMol.ase_atoms[genMol_subset_component_inds].copy() 
        tmp_ref_comp = compareMol.ase_atoms[compareMol_subset_component_inds].copy() 

        flag_struct = False

        if np.any(sorted(tmp_self_comp.get_atomic_numbers()) != sorted(tmp_ref_comp.get_atomic_numbers())):
            print('Warning cores not matched at graph depth!')
            print('Can re-attempt. Not doing that now.')
            print(tmp_self_comp.get_chemical_formula())
            print(tmp_ref_comp.get_chemical_formula())
            outcore = tmp_self_comp
            rmsd_loss_core = 1000
            rmsd_loss_full = 1000
            flag_struct = True
        else:
            # Center on metal atom
            tmp_self_comp.set_positions(tmp_self_comp.positions - genMol.ase_atoms[genMol_metalind].position)
            tmp_ref_comp.set_positions(tmp_ref_comp.positions - compareMol.ase_atoms[compareMol_metalind].position)
            genMol.ase_atoms.set_positions(genMol.ase_atoms.positions - genMol.ase_atoms[genMol_metalind].position)
            compareMol.ase_atoms.set_positions(compareMol.ase_atoms.positions - compareMol.ase_atoms[compareMol_metalind].position)
            
            # Sample random rotations to find best starting assignment point.
            best = np.inf
            for _ in range(sample):
                q = Rot.random()
                calc_test_comp = copy.deepcopy(tmp_self_comp)
                calc_test_comp.set_positions(q.apply(calc_test_comp.positions))
                rmsd_core, _, _ = permute_align(tmp_ref_comp, calc_test_comp, maxiter=maxiter)
                rmsd_mirror, _, _ = mirror_align(tmp_ref_comp, calc_test_comp, maxiter=maxiter)
                if rmsd_mirror < rmsd_core:
                    rmsd_core = rmsd_mirror
                if rmsd_core < best:
                    saveq = q
                    best = rmsd_core
                
            tmp_self_comp.set_positions(saveq.apply(tmp_self_comp.positions))
            genMol.ase_atoms.set_positions(saveq.apply(genMol.ase_atoms.positions))
            
            rmsd_core, r, outcore = permute_align(tmp_ref_comp, tmp_self_comp, maxiter=maxiter)
            rmsd_mirror, r_mirror, moutcore = mirror_align(tmp_ref_comp, tmp_self_comp, maxiter=maxiter)

            if rmsd_mirror < rmsd_core: # Pick the better one!
                rmsd_core = rmsd_mirror
                outcore = moutcore
                r = r_mirror
                newposits = genMol.ase_atoms.positions
                newposits[:,0] = -newposits[:,0] # Mirror across x axis to replicate mirror in permute
                genMol.ase_atoms.set_positions(newposits)
                    
            rmsd_loss_core = rmsd_core
            tmp_posits = genMol.ase_atoms.positions
            tmp_posits = r.apply(tmp_posits)
            genMol.ase_atoms.set_positions(tmp_posits)
        ## Do permutation mapping to estimate full loss given the rotation to match the core.
            rmsd_loss_full, _, _ = permute_align(copy.deepcopy(compareMol.ase_atoms), copy.deepcopy(genMol.ase_atoms), 
                                                    maxiter=1, tol=1e-6, in_place=True)
        
    if rmsd_type == 'simple':
        # Returns per-atom RMSD for ideal/rotation/translation overlap.
        if return_structures:
            return (rmsd_loss_core, 
                    rmsd_loss_full, 
                    compareMol.write_mol2('refmol.mol2',writestring=True), 
                    genMol.write_mol2('aligned.mol2',writestring=True),
                    convert_ase_xyz(tmp_ref_comp),
                    convert_ase_xyz(outcore),
                    flag_struct)
        else:
            return (rmsd_loss_core, rmsd_loss_full, flag_struct)
    else:
        print('Not yet implemented.')
        return None


def calc_rmsd_atypes(genMol, compareMol, sample=300, atom_types=None,
            rmsd_type='simple'): 
    """calc_rmsd 
    Calculate the rmsd by different methods for this molecule compared to another.

    Parameters
    ----------
    genMol : str/Molecule
        molecule generated to rotate to reference.
    compareMol : str/Molecule
        molecule to compare to (reference molecule)
    rmsd_type : str, optional
        which type of rmsd to calculate , by default 'simple', possible 'edge'
    sample : int, optional
        number of random rotations to sample initially to ensure best mapping between cores, by default 30
    atom_types: list, optional
        which atom types to consider for alignment - will default to full alignment, default None

    Returns
    -------
    rmsd : float
        RMSD per atom comparison between the molecules.
    """

    genMol = convert_io_molecule(genMol)
    compareMol = convert_io_molecule(compareMol)

    # Check that these are stoichiometrically identical molecules.
    # Set ordering to be identical based on canonical labels.
    if atom_types is None:
        compareMol_subset_component_inds = np.array([i for i,x in enumerate(compareMol.ase_atoms.get_chemical_symbols())])
        genMol_subset_component_inds = np.array([i for i,x in enumerate(genMol.ase_atoms.get_chemical_symbols())])
    elif isinstance(atom_types,str): 
        if atom_types == 'metals':
            compareMol_subset_component_inds = np.array([i for i,x in enumerate(compareMol.ase_atoms.get_chemical_symbols()) if x in io_ptable.all_metals])
            genMol_subset_component_inds = np.array([i for i,x in enumerate(genMol.ase_atoms.get_chemical_symbols()) if x in io_ptable.all_metals])
        elif atom_types == 'heavy_atoms':
            compareMol_subset_component_inds = np.array([i for i,x in enumerate(compareMol.ase_atoms.get_chemical_symbols()) if x != "H"])
            genMol_subset_component_inds = np.array([i for i,x in enumerate(genMol.ase_atoms.get_chemical_symbols()) if x != "H"])
        elif atom_types == 'heavy_metals':
            compareMol_subset_component_inds = np.array([i for i,x in enumerate(compareMol.ase_atoms.get_chemical_symbols()) if x in io_ptable.heavy_metals])
            genMol_subset_component_inds = np.array([i for i,x in enumerate(genMol.ase_atoms.get_chemical_symbols()) if x in io_ptable.heavy_metals])
        else:
            raise ValueError('I do not recognize this keyword "{}", please choose from "metals" or "heavy_atoms" or "heavy_metals"'.format(atom_types))
    else:
        compareMol_subset_component_inds = np.array([i for i,x in enumerate(compareMol.ase_atoms.get_chemical_symbols()) if x in atom_types])
        genMol_subset_component_inds = np.array([i for i,x in enumerate(genMol.ase_atoms.get_chemical_symbols()) if x in atom_types])

    # Pull out selected atoms from molecule up to depth coresize graph hops for matching to reference.
    tmp_self_comp = genMol.ase_atoms[genMol_subset_component_inds].copy() 
    tmp_ref_comp = compareMol.ase_atoms[compareMol_subset_component_inds].copy() 

    flag_struct = False

    if np.any(sorted(tmp_self_comp.get_atomic_numbers()) != sorted(tmp_ref_comp.get_atomic_numbers())):
        print('Warning subset stoichiometry not matched!')
        print('Can re-attempt. Not doing that now.')
        print(tmp_self_comp.get_chemical_formula())
        print(tmp_ref_comp.get_chemical_formula())
        outcore = tmp_self_comp
        rmsd_loss_core = 1000
        rmsd_loss_full = 1000
        flag_struct = True
    else:
        # Center on metal atom
        tmp_self_comp.set_positions(tmp_self_comp.positions - tmp_self_comp.get_positions().mean(axis=0))
        tmp_ref_comp.set_positions(tmp_ref_comp.positions - tmp_ref_comp.get_positions().mean(axis=0))
        genMol.ase_atoms.set_positions(genMol.ase_atoms.positions - tmp_self_comp.get_positions().mean(axis=0))
        compareMol.ase_atoms.set_positions(compareMol.ase_atoms.positions - tmp_ref_comp.get_positions().mean(axis=0))
        
        # Sample random rotations to find best starting assignment point.
        best = np.inf
        for _ in range(sample):
            q = Rot.random()
            calc_test_comp = copy.deepcopy(tmp_self_comp)
            calc_test_comp.set_positions(q.apply(calc_test_comp.positions))
            rmsd_core, _, _ = permute_align(tmp_ref_comp, calc_test_comp)
            rmsd_mirror, _, _ = mirror_align(tmp_ref_comp, calc_test_comp)
            if rmsd_mirror < rmsd_core:
                rmsd_core = rmsd_mirror
            if rmsd_core < best:
                saveq = q
                best = rmsd_core
            
        tmp_self_comp.set_positions(saveq.apply(tmp_self_comp.positions))
        genMol.ase_atoms.set_positions(saveq.apply(genMol.ase_atoms.positions))
        
        rmsd_core, r, outcore = permute_align(tmp_ref_comp, tmp_self_comp)
        rmsd_mirror, r_mirror, moutcore = mirror_align(tmp_ref_comp, tmp_self_comp)

        if rmsd_mirror < rmsd_core: # Pick the better one!
            rmsd_core = rmsd_mirror
            outcore = moutcore
            r = r_mirror
            newposits = genMol.ase_atoms.positions
            newposits[:,0] = -newposits[:,0] # Mirror across x axis to replicate mirror in permute
            genMol.ase_atoms.set_positions(newposits)
                
        rmsd_loss_core = rmsd_core
        tmp_posits = genMol.ase_atoms.positions
        tmp_posits = r.apply(tmp_posits)
        genMol.ase_atoms.set_positions(tmp_posits)
    ## Do permutation mapping to estimate full loss given the rotation to match the core.
        if compareMol.ase_atoms.get_chemical_formula() == genMol.ase_atoms.get_chemical_formula():
            rmsd_loss_full, _, _ = permute_align(copy.deepcopy(compareMol.ase_atoms), copy.deepcopy(genMol.ase_atoms), 
                                                    maxiter=1, tol=1e-6, in_place=True)
        else:
            rmsd_loss_full = None
        
    if rmsd_type == 'simple':
        # Returns per-atom RMSD for ideal/rotation/translation overlap.
        return (rmsd_loss_core, 
                rmsd_loss_full, 
                compareMol.write_mol2('refmol.mol2',writestring=True), 
                genMol.write_mol2('aligned.mol2',writestring=True),
                convert_ase_xyz(tmp_ref_comp),
                convert_ase_xyz(outcore),
                flag_struct)
    else:
        print('Not yet implemented.')
        return None


"""
Core geometry information mining.

Geomety class contains all the information of possible cores and ligands mapping to cores.

Routines for mapping user-defined cores/ligands encoded here!

Developed by Michael Taylor and Jan Janssen
"""

import itertools
import numpy as np
import pandas as pd
import os
import architector.io_ptable as io_ptable
import architector.geometries as geo
import warnings
import architector.io_molecule as io_molecule

warnings.filterwarnings('error')

np.seterr(all='warn')


def get_lig_ref_df():
    """get_lig_ref_df 
    Pull the ligand angle references.

    Returns
    -------
    ref_df : dict
        dictionary with {'type':angles(List)}
    """
    filepath = os.path.abspath(os.path.join(__file__, "..", "data", "ligtype_angle_reference.csv"))
    ref_df = pd.read_csv(filepath)
    return ref_df


def get_angle(coord1, coord2, coord3):
    """get_angle 

    Parameters
    ----------
    coord1 : list
        coordinates of point 1
    coord2 : list
        coordinates of point 2
    coord3 : list
        coordinates of point 3

    Returns
    -------
    angle : float
        angle between 1-2-3 (2 at center)
    """
    if isinstance(coord1, list):
        coord1 = np.array(coord1)
    if isinstance(coord2, list):
        coord2 = np.array(coord2)
    if isinstance(coord3, list):
        coord3 = np.array(coord3)
    v1 = coord1-coord2
    v2 = coord3-coord2
    try:
        angle = np.degrees(np.arccos(np.dot(v1, v2)/(np.linalg.norm(v1)*np.linalg.norm(v2))))
        if np.isnan(angle):
            if np.any(np.isclose(v1/np.linalg.norm(v1) + v2/np.linalg.norm(v2),0)):
                return 180.0
            else:
                return 0.0
        return angle
    except Warning: # Catch cases where angles are too close (parallel/antiparallel)
        if np.any(np.isclose(v1/np.linalg.norm(v1) + v2/np.linalg.norm(v2),0)):
            return 180.0
        else:
            return 0.0
    

def check_intercalation(coordat_list, all_coordats):
    """check_intercalation 
    Check for any intercalating points between selected indices

    Parameters
    ----------
    coordat_list : list
        list of coordinating atoms selected for specific geometry
    all_coordats : list
        all possible coordinating atoms for specific geometry

    Returns
    -------
    intercalating : bool
        whether or not any of the remaining coord atoms at the surface of the metal
        fall between the potential selected coordinating atoms for the ligand
    """
    metal_coords = [0.0, 0.0, 0.0]
    coordats = np.array(coordat_list)
    if len(coordats) > 1:
        all_coordats = np.array(all_coordats)
        test_coordats = []
        for coord in all_coordats:  # Remove matched coordatoms from full list.
            good = True
            for coordat in coordats:
                if np.linalg.norm(coord-coordat) < 1e-12:
                    good = False
            if good:
                test_coordats.append(coord)
        test_coordats = np.array(test_coordats)
        intercalating = False
        for pair in itertools.combinations(coordats, 2):  # Take pairs of coord atoms
            if get_angle(pair[0], metal_coords, pair[1]) < 121:  # If the angle is less than 121 check for intercalation
                midvect = (pair[0] + pair[1])  # Take the midpoint vector between the two cord points
                for coordat in test_coordats:  # Test the remaining coordintion sites
                    if get_angle(midvect, metal_coords, coordat) < 30:
                        # If any of the points are within 30 degrees of the midpoint
                        # Flag as intercalating
                        intercalating = True
    else:
        intercalating = False
    return intercalating

denticity_combinations_dict = {0:1, 1:2, 3:3, 6:4, 10:5, 15:6, 21:7, 28:8, 36:9}

def calc_angles_struct(structure):
    """ Read in structure to assign ligand to right shape

    structure : mol2string
    """
    if isinstance(structure,list):
        lig_angles = np.array(structure)[np.argsort(angs)[::-1]]
    else:
        mol = io_molecule.convert_io_molecule(structure)
        if len(mol.graph) == 0:
            print('Creating imputed molecular graph! May be untrustworthy.')
            mol.create_BO_dict()
        mets = mol.find_metals()
        lig_angles = []
        if len(mets) == 1:
            coordats = np.nonzero(mol.graph[mets[0]])[0]
            if len(coordats) == 1:
                lig_angles = []
            else:
                coords = mol.ase_atoms.get_positions()
                angs = [
                    get_angle(coords[x[0]],coords[mets[0]],coords[x[1]]) for x in itertools.combinations(coordats,2)
                    ]
                angs = np.array(angs)[np.argsort(angs)[::-1]] # Add angles
                lig_angles += angs.tolist() # Add sorted angles as features
        else:
            print('Warning: User ligand input without metal for refernce on interatomic angles. \
                    Please pass a structure with a metal for user ligand generation.')
    lig_angles += [0.0] * (36-len(lig_angles)) # Pad with zeros
    n_ca_m_ca_angles = len(np.nonzero(lig_angles)[0])
    denticity = denticity_combinations_dict[n_ca_m_ca_angles]
    return {'user_lig':np.array(lig_angles)}, denticity

class Geometries:
    """ geometries Dataclass containing information and corecoordLists for all
    most frequent molecular geometries:
    https://en.wikipedia.org/wiki/Capped_square_antiprismatic_molecular_geometry 
    """
    def __init__(self, usercore=False):
        """__init__ 
        """
        self.geometry_dict = {
            f: getattr(geo, f)
            for f in dir(geo)
            if not f.endswith("__")
        }

        # Add in user's core information
        if isinstance(usercore,(list,np.ndarray)):
            self.geometry_dict.update({'user_core':usercore})

        self.geo_cn_dict = {
            k: len(v)
            for k, v in self.geometry_dict.items()
        }

        self.cn_geo_dict = {
            k: []
            for k in set(self.geo_cn_dict.values())
        }
        for k, v in self.geo_cn_dict.items():
            self.cn_geo_dict[v].append(k)

        self.ambiguous_goes = [
            'bicapped_trigonal_prismatic',
            'capped_trigonal_prismatic',
            'capped_octahedral',
            'capped_square_antiprismatic'
        ]

        self.trans_geos = [
            'linear',
            't_shaped',
            'square_planar',
            'seesaw',
            'square_pyramidal',
            'trigonal_bipyramidal',
            'octahedral',
            'pentagonal_bipyramidal',
            'hexagonal_bipyramidal',
            'axial_bicapped_trigonal_prismatic',
            'axial_bicapped_hexagonal_planar',
            'penta_bi_capped_pyramidal',
            'senrag_comp_1'
        ]

        self.liglist_geo_map_dict = None

    def rescale_refgeos(self, core_atom, rcovmetal=None):
        """rescale_refgeos 
        rescale the geometry dictionary to closer match the actual atomic radii
        matters for non-perfectly symmetric core geometries!

        Parameters
        ----------
        core_atom : str
            chemical symbol of core atom
        rcovmetal : float, optional
            Covalent radii of the metal, defaults to values in io_ptable.rcov1
        """
        if isinstance(rcovmetal,float):
            interatomic_dist_ideal = rcovmetal + io_ptable.rcov1[io_ptable.elements.index('O')]
        else:
            interatomic_dist_ideal = \
                io_ptable.rcov1[io_ptable.elements.index(core_atom)] + io_ptable.rcov1[io_ptable.elements.index('O')]
        geodict_copy = self.geometry_dict.copy()
        for key in self.geometry_dict.keys():
            geo_dict = self.geometry_dict[key]
            dists = np.linalg.norm(geo_dict, axis=1)
            avg_dist = np.mean(dists)
            re_scale = interatomic_dist_ideal/avg_dist
            scaled_geo = []
            for xyz in geo_dict:
                out = [xyz[0]*re_scale, xyz[1]*re_scale, xyz[2]*re_scale]
                scaled_geo.append(out)
            geodict_copy.update({key: scaled_geo})
        self.geometry_dict = geodict_copy

    def get_lig_ref_inds_dict(self, core_atom, coreTypes, userlig=False, rcovmetal=None): 
        """get_lig_ref_inds_dict 
        Map the reference ligand angle information onto possible
        core geometries!

        Parameters
        ----------
        core_atom : str
            metal atomic symbol
        coreTypes : list
            list of core coordination types (e.g. ['octahedral','tetrahedral'...])
        userlig : bool/list/molecule.
            list of reference ligands
        rcovmetal : float
            radii to be imposed for metal
        """
        df = get_lig_ref_df()  # Get ligand angle reference information

        # Translate reference data into ligands info here.
        self.ligType_cn_dict = dict()
        for i,row in df.iterrows():
            self.ligType_cn_dict.update({row['ligtype']:int(row['denticity'])})
        
        self.cn_ligType_dict = {
            k: []
            for k in set(self.ligType_cn_dict.values())
        }
        for k, v in self.ligType_cn_dict.items():
            self.cn_ligType_dict[v].append(k)

        self.rescale_refgeos(core_atom,rcovmetal=rcovmetal)
        # Add userlig information here - handle either reference ligand/metal geometry
        if not isinstance(userlig,bool):
            userlig, denticities = calc_angles_struct(userlig)
            count = 0
            for key,val in userlig.items():
                df.loc[len(df),:] = list(val)
                self.ligType_cn_dict.update({'user_lig'+str(count):denticities[count]})
                tlist = self.cn_ligType_dict[denticities[count]]
                tlist.append('user_lig'+str(count))
                self.cn_ligType_dict.update({denticities[count]:tlist})
                count += 1
        mean_angle_labels = [x for x in df.columns.values if '_mean' in x]  # Mean angle!
        metal_coords = [0.0, 0.0, 0.0]
        outdict_rows = []
        labels = []
        possible_geos = coreTypes

        for i, row in df.iterrows():
            outdict = dict()
            outdict_losses = dict()
            labels.append(row['ligtype'])
            cn_min = row['denticity']
            mean_angles = row[mean_angle_labels].values
            max_angles = mean_angles + np.ones(36)*30  # Use looser criterion to get more points.
            min_angles = mean_angles - np.ones(36)*30
            for core_type in possible_geos:  # Define possible coordination environments and selected indices
                # Rule out pairs or sets where additional points fall between the selected indices.
                coordat_locs = self.geometry_dict[core_type]
                all_coordat_inds = [x for x in itertools.combinations(range(len(coordat_locs)), cn_min)]
                all_coordat_combs = []
                for x in all_coordat_inds:
                    tmp = []
                    for k in x:
                        tmp.append(coordat_locs[k])
                    all_coordat_combs.append(tmp)
                for j, coordat_comb in enumerate(all_coordat_combs):
                    angs = [
                        get_angle(x[0], metal_coords, x[1]) for x in itertools.combinations(coordat_comb, 2)
                        ]
                    angs = np.array(angs)[np.argsort(angs)[::-1]]  # sort angles largest first!
                    if cn_min < 9:  # pad angles with zeros
                        angs = np.pad(angs, (0, 36-len(angs)), 'constant')
                    if np.all(angs <= max_angles) & np.all(angs >= min_angles):
                        loss = np.mean(np.abs(angs - mean_angles))  # MAE
                        # Check for intercalating coordination points!!! -> don't use these!!!
                        intercalating = check_intercalation(coordat_comb, coordat_locs)
                        if (core_type in outdict) and (not intercalating):
                            temp = outdict[core_type].copy()
                            temp.append(all_coordat_inds[j])
                            outdict.update({core_type: temp})
                            tmp_losses = outdict_losses[core_type].copy()
                            tmp_losses.append(loss)
                            outdict_losses.update({core_type: tmp_losses})  # Save losses
                        elif not intercalating:
                            outdict.update({core_type: [all_coordat_inds[j]]})
                            outdict_losses.update({core_type: [loss]})
                # Sort outdict by losses
                for key in outdict.keys():
                    tmp_inds = outdict[key].copy()
                    tmp_losses = outdict_losses[key]
                    order = np.argsort(tmp_losses)  # Smallest loss first
                    out = np.array(tmp_inds)[order]
                    out = out.tolist()
                    outdict[key] = out
            outdict_rows.append(outdict)

        for dent in [10,11,12]: # For higher denticity ligands set them to all possible combinations of locations
            labels.append(str(dent))
            cn_min=dent
            outdict = dict()
            for core_type in possible_geos:
                coordat_locs = self.geometry_dict[core_type]
                all_coordat_inds = [x for x in itertools.combinations(range(len(coordat_locs)), cn_min)]
                outdict.update({core_type:all_coordat_inds})
            outdict_rows.append(outdict)

        total_dict = dict(zip(labels, outdict_rows))
        # Assign sandwich to tri_fac geometries!
        total_dict['sandwich'] = total_dict['tri_fac']
        self.liglist_geo_map_dict = total_dict


def calc_all_coord_atom_angles(core_geo):
    """calc_all_coord_atom_angles 
    Get all of the L-M-L angles for a core geometry. 

    Parameters
    ----------
    core_geo : np.ndarray
        core geometry with coordinates of all coordinating atoms.

    Returns
    -------
    np.ndarray
        sorted array of L-M-L angles
    """
    origin = np.array((0,0,0))
    angles = []
    for a,b in itertools.combinations(core_geo,2):
        angles.append(get_angle(a,origin,b))
    return np.array(sorted(angles))
        

def classify_metal_geo_type(mol2string,metal_ind=None):
    """classify_metal_geo_type calculate the actual geometry of the metal centers

    Parameters
    ----------
    mol2string : str
        mol2string with metals
    metal_ind : int or list, optional
        the metal indices, by default None

    Returns
    -------
    metal_center_geos : dict or list
        metal center geometries present in the mol2string.
    """
    molecule = io_molecule.convert_io_molecule(mol2string)
    if metal_ind is None:
        metal_ind = [i for i,x in enumerate(molecule.ase_atoms.get_chemical_symbols()) if (x in io_ptable.all_metals)]
    elif isinstance(metal_ind,int):
        metal_ind = [metal_ind]
    elif isinstance(metal_ind,(np.ndarray,list)):
        metal_ind = metal_ind 
    else:
        raise ValueError('Unrecognized type for metal_ind')
    geo_dict = Geometries()
    if len(metal_ind) == 0:
        raise ValueError('No metal or ind passed in this molecule.')
    elif len(metal_ind) > 1: # Look at every metal center
        metal_center_geos = []
        for metal_indx in metal_ind: 
            tmpdict = dict()
            neighs = np.nonzero(np.ravel(molecule.graph[metal_indx]))[0]
            if (len(neighs) < 13) and (len(neighs) != 11):
                coord_at_positions = molecule.ase_atoms.positions[neighs] - molecule.ase_atoms.positions[metal_indx]
                act_geo_vect = calc_all_coord_atom_angles(coord_at_positions)
                ref_geo_labels = geo_dict.cn_geo_dict[len(neighs)]
                ref_geos = [calc_all_coord_atom_angles(geo_dict.geometry_dict[x]) for x in ref_geo_labels]
                mae_losses = [np.mean(np.abs(act_geo_vect - x)) for x in ref_geos] # Calc MAE loss between interatomic angles
                sort_order = np.argsort(mae_losses)
                m_geo_type = ref_geo_labels[np.argmin(mae_losses)]
                tmpdict['metal'] = molecule.ase_atoms.get_chemical_symbols()[metal_indx]
                tmpdict['metal_ind'] = metal_indx
                tmpdict['metal_geo_type'] = m_geo_type
                tmpdict['mae_angle_loss'] = mae_losses[np.argmin(mae_losses)]
                if len(sort_order) > 1:
                    tmpdict['confidence'] = 1 - tmpdict['mae_angle_loss'] / mae_losses[sort_order[1]]
                else:
                    tmpdict['confidence'] = 1
                tmpdict['classification_dict'] = {ref_geo_labels[i]:mae_losses[i] for i in sort_order}
            else: 
                tmpdict['metal'] = molecule.ase_atoms.get_chemical_symbols()[metal_indx]
                tmpdict['metal_ind'] = metal_indx
                tmpdict['metal_geo_type'] = len(neighs)
            metal_center_geos.append(tmpdict)
    else: # Just calculate geometry for one metal center.
        metal_center_geos = dict()
        metal_indx = metal_ind[0]
        neighs = np.nonzero(np.ravel(molecule.graph[metal_indx]))[0]
        if len(neighs) < 13 and (len(neighs) != 11):
            coord_at_positions = molecule.ase_atoms.positions[neighs] - molecule.ase_atoms.positions[metal_indx]
            act_geo_vect = calc_all_coord_atom_angles(coord_at_positions)
            ref_geo_labels = geo_dict.cn_geo_dict[len(neighs)]
            ref_geos = [calc_all_coord_atom_angles(geo_dict.geometry_dict[x]) for x in ref_geo_labels]
            mae_losses = [np.mean(np.abs(act_geo_vect - x)) for x in ref_geos]
            sort_order = np.argsort(mae_losses)
            m_geo_type = ref_geo_labels[np.argmin(mae_losses)]
            metal_center_geos['metal'] = molecule.ase_atoms.get_chemical_symbols()[metal_indx]
            metal_center_geos['metal_ind'] = metal_indx
            metal_center_geos['metal_geo_type'] = m_geo_type
            metal_center_geos['mae_angle_loss'] = mae_losses[np.argmin(mae_losses)]
            if len(sort_order) > 1:
                metal_center_geos['confidence'] = (mae_losses[sort_order[1]]-metal_center_geos['mae_angle_loss'])/metal_center_geos['mae_angle_loss']
            else:
                metal_center_geos['confidence'] = 1
            metal_center_geos['classification_dict'] = {ref_geo_labels[i]:mae_losses[i] for i in sort_order}
        else: 
            metal_center_geos['metal'] = molecule.ase_atoms.get_chemical_symbols()[metal_indx]
            metal_center_geos['metal_ind'] = metal_indx
            metal_center_geos['metal_geo_type'] = len(neighs)
    return metal_center_geos


"""
IO module for open babel.

Developed by Dan Burril and Michael Taylor
"""

# Imports
import ase
import numpy as np
import architector
import architector.io_ptable as io_ptable
import warnings

from ase.io import read
from ase import units
from openbabel import openbabel as ob
from openbabel import pybel
from io import StringIO
from scipy.sparse import csgraph
import numpy as np
from pynauty import Graph as pnGraph
from pynauty import canon_label

ob_log_handler = ob.OBMessageHandler()
ob_log_handler.SetOutputLevel(0) # Set warnings to only critical.

warnings.filterwarnings('ignore') # Supress warnings.

# Functions
def smiles2xyz(smilesStr,addHydrogens=True):
    '''smiles2xyz
    Convert smiles string to xyz string.

    smilesStr : str
        smiles to convert to xyz file
    addHydrogens : bool
        whether to add hydrogens or not?!?, default True
    '''

    # Set up conversion
    obConversion = ob.OBConversion()
    obConversion.SetInAndOutFormats("smi", "xyz")
    mol = ob.OBMol()
    obConversion.ReadString(mol, smilesStr)

    # Add hydrogens
    if (addHydrogens == True):
        mol.AddHydrogens()

    # Generate 3d structure
    builder = ob.OBBuilder()
    builder.Build(mol)

    # Set up force field
    FF = ob.OBForceField.FindForceField("MMFF94")
    FF.Setup(mol)

    # Optimize energy
    FF.ConjugateGradients(2000,1e-6)
    FF.GetCoordinates(mol)

    # Perform conversion
    obConversion.Convert()

    return obConversion.WriteString(mol).strip()

def smiles2Atoms(smilesStr,addHydrogens=True):
    """smiles2Atoms
    Convert SMILES string to Atoms object.

    Parameters
    ----------
    smilesStr : str
        smiles string
    addHydrogens : bool, optional
       add hydrogens to the molecule?, by default True

    Returns
    -------
    ats : ase.atoms.Atoms
        ase atoms of the passed smiles string string
    """

    # Variables
    symList = []        # Hold atomic symbols
    posList = []        # Hold atomic positions

    # Convert smiles to xyz
    xyzStr = smiles2xyz(smilesStr,addHydrogens=addHydrogens).split("\n")[2:]

    # Get symbols and positions
    for line in xyzStr:
        # Format line
        line = line.strip().split()

        # Get information
        symList.append(line[0])
        pos = [float(val) for val in line[1:]]
        posList.append(pos) 

    # Create Atoms object
    ats = ase.Atoms("".join(symList),positions=posList)

    # Center at COM
    ats.translate(-ats.get_center_of_mass())
    return ats

def get_OBMol_coords_anums_graph(OBMol, return_coords=True, get_types=False):
    """get_OBMol_coords_anums_graph 
    Mine obmol structure for molecular graph, atomic numbers, and coordinates.

    Parameters
    ----------
    OBMol : OBMol
        OBMol object with information
    return_coords : bool, optional
        Return coordinates, default True
    get_types : bool, optional
        Return atom types instead of numbers, default False

    Returns
    -------
    coords : np.ndarray
        NX3 array of xyz coordinates
    anums : np.ndarray
        N array of atomic numbers
    outgraph : np.ndarray
        NXN array of molecular graph
    """
    
    anums = []
    for i, atom in enumerate(ob.OBMolAtomIter(OBMol)):
        if not get_types:
            anums.append(atom.GetAtomicNum())
        else:
            anums.append(atom.GetType())
    coords = []
    if return_coords:
        for i, atom in enumerate(ob.OBMolAtomIter(OBMol)):
            vect = (atom.GetX(),atom.GetY(),atom.GetZ())
            coords.append(vect)
    outgraph = np.zeros((len(anums),len(anums)))
    for obbond in ob.OBMolBondIter(OBMol):
        outgraph[obbond.GetBeginAtomIdx()-1,obbond.GetEndAtomIdx()-1] = 1
        outgraph[obbond.GetEndAtomIdx()-1,obbond.GetBeginAtomIdx()-1] = 1
    coords = np.array(coords)
    return coords, anums, outgraph

def get_OBMol_bo_dict_atom_types(OBMol,metal_passed=True):
    """get_OBMol_coords_anums_graph 
    Mine obmol structure for BO dict and atom types

    Parameters
    ----------
    OBMol : OBMol
        OBMol object with information

    Returns
    -------
    bo_dict : dict
        Bond order dictionary with {(i,j):order} format
        Bond i,j 1-indexed with 1 always corresponding to the metal atom.
    atypes : list
        Atom types of the non-metal atoms
    """
    
    atypes = []
    ttab = ob.OBTypeTable()
    ttab.SetFromType('INT')
    ttab.SetToType('SYB')
    natoms = OBMol.NumAtoms()
    for i, atom in enumerate(ob.OBMolAtomIter(OBMol)):
        atypes.append(ttab.Translate(atom.GetType()))
    atypes = atypes[:-1] # Get rid of last atom (will always be metal)
    bo_dict = dict()
    for obbond in ob.OBMolBondIter(OBMol):
        bo = obbond.GetBondOrder()
        i = obbond.GetBeginAtomIdx()
        j = obbond.GetEndAtomIdx()
        if metal_passed:
            if i == natoms:
                i = 1
            else:
                i = 1 + i
            if j == natoms:
                j = 1
            else:
                j = 1 + j
            if i > j:
                bo_dict.update({(j,i):bo})
            else:
                bo_dict.update({(i,j):bo})
        else:
            if i > j:
                bo_dict.update({(j,i):bo})
            else:
                bo_dict.update({(i,j):bo})
    return bo_dict, atypes

def check_mmff_okay(OBMol):
    """check_mmff_okay 
    check if mmff94 applicable to chemistry

    Parameters
    ----------
    OBMol : OBmol
        openbabel of the molecule

    Returns
    -------
    mmff94good : bool
        whether this structure is okay to evaluate with mmff94.
    """
    mmff94_good_elements = ['C','H','N','O','F','Si','P','S','Br', 'Cl', 'I']
    mmff94_good_anums = [io_ptable.elements.index(x) for x in mmff94_good_elements]
    mmff94good = True
    for i, atom in enumerate(ob.OBMolAtomIter(OBMol)):
        if atom.GetAtomicNum() not in mmff94_good_anums:
            mmff94good = False
    return mmff94good
            

def Neutralize(OBmol):
    """Neutralize 
    apply openbabel neutralization procedure
    Useful for binding energies.

    Parameters
    ----------
    OBmol : openbabel.OBmol
       molecule to neutralize

    Returns
    -------
    changed : bool 
        wheter or not the molecule was changed
    """
    neutralize = ob.OBOp.FindType("neutralize")
    option = ""
    changed = neutralize.Do(OBmol, option)
    return changed


def get_obmol_smiles(smilesStr,
                    addHydrogens=True,
                    neutralize=False,
                    build=True,
                    functionalizations=None):
    """get_obmol_smiles
    convert smiles to OBmol instance

    Parameters
    ----------
    smilesStr : str
        smiles string
    addHydrogens : bool, optional
        Use add hydrogens function, by default True
    neutralize : bool, optional
        Whether to add/remove protons to make a neutral ligand, default False
    build : bool, optional
        whether to build into 3D, default True
    functionalizations : list, optional
        list of functionalization dictionaries, default None

    Returns
    -------
    OBmol : OBMol
        Built openbabel molecule
    """
    # Set up conversion
    obConversion = ob.OBConversion()
    obConversion.SetInAndOutFormats("smi", "xyz")
    OBmol = ob.OBMol()
    obConversion.ReadString(OBmol, smilesStr)

    if (functionalizations is not None):
        for fg in functionalizations:
            OBmol = functionalize(OBmol, 
                                  functional_group=fg['functional_group'],
                                  smiles_inds=fg['smiles_inds'])

    if neutralize:
        _ = Neutralize(OBmol)

    # Add hydrogens
    if (addHydrogens == True):
        OBmol.AddHydrogens()

    if build:
        OBmol = build_3D(OBmol,addHydrogens=addHydrogens)
        return OBmol
    else:
        return OBmol


def get_smiles_obmol(OBmol,canonicalize=False):
    """get_obmol_smiles
    convert smiles to OBmol instance

    Parameters
    ----------
    OBmol : ob.OBMol
        Built openbabel molecule
    canonicalize : bool, optional
        canonicalize the smiles?
    
    Returns
    -------
    smiles : str
        smiles of the OBmol
    """
    # Set up conversion
    obConversion = ob.OBConversion()
    obConversion.SetOutFormat('smi')
    if canonicalize:
        obConversion.SetOutFormat('can')
    smiles = obConversion.WriteString(OBmol).split()[0]
    return smiles


def canonicalize_smiles(insmiles):
    """get_obmol_smiles
    convert smiles to OBmol instance

    Parameters
    ----------
    insmiles : str
        smiles of the moleucle
    
    Returns
    -------
    can_smiles : str
       canonicalized smiles of the molecule
    """
    # Set up conversion
    obConversion = ob.OBConversion()
    obConversion.SetInFormat('smi')
    obConversion.SetOutFormat('can')
    OBmol = ob.OBMol()
    obConversion.ReadString(OBmol, insmiles)
    can_smiles = obConversion.WriteString(OBmol).split()[0]
    return can_smiles


def build_3D(OBmol,addHydrogens=True):
    """build_3D take a 2D OBmol structure and build to 3D

    Parameters
    ----------
    OBmol : ob.OBMol
        2D OBmol structure
    addHydrogens : bool, optional
        whether to add hydrogens before building, by default True

    Returns
    -------
    OBmol
        3D Built OBmol structure
    """
    # Generate 3d structure
    if (addHydrogens == True):
        OBmol.AddHydrogens()
    builder = ob.OBBuilder()
    builder.Build(OBmol)

    mmff94_ok = check_mmff_okay(OBmol)

    # Set up force field
    if mmff94_ok:
        FF = ob.OBForceField.FindForceField("mmff94")
    else:
        FF = ob.OBForceField.FindForceField('UFF')
    FF.Setup(OBmol)

    # Optimize energy
    FF.ConjugateGradients(2000,1e-6)
    FF.GetCoordinates(OBmol)
    return OBmol


def generate_obmol_conformers(smiles, rmsd_cutoff=0.4, conf_cutoff=3000, energy_cutoff=50.0, 
        confab_verbose = False, output_format='mol2', neutralize=False, functionalizations=None):
    """generate_obmol_conformers 
    generate conformers with openbabel for given smiles
    using confab conformer generation routine
    O'Boyle NM, Vandermeersch T, Flynn CJ, Maguire AR, Hutchison GR. Confab 
    - Systematic generation of diverse low-energy conformers. Journal of Cheminformatics. 
    2011;3:8. doi:10.1186/1758-2946-3-8.

    Parameters
    ----------
    smiles : str
        smiles to generate conformers 
    rmsd_cutoff : float, optional
        cutoff for how similar conformers, by default 0.4
    conf_cutoff : int, optional
        total number of conformers to generate, by default 3000
    energy_cutoff : float, optional
        how similar in energy, by default 50.0
    confab_verbose : bool, optional
        give more detailed output, by default False
    output_format : str, optional
        which format to output , by default 'mol2'
    neutralize : bool, optional
        neutralize smiles?, by default False
    functionalizations : dict, optional
        add functionalizations?, by default None

    Returns
    -------
    output_strings : list (str)
        list of conformers generated as whatever format desired
    """
    obmol = get_obmol_smiles(smiles,
                             neutralize=neutralize,
                             functionalizations=functionalizations)
    mmff94_ok = check_mmff_okay(obmol)
    if mmff94_ok:
        FF = ob.OBForceField.FindForceField("MMFF94")
        FF.Setup(obmol) # Make sure setup works OK
    else:
        FF = ob.OBForceField.FindForceField("MMFF94")
        FF.Setup(obmol) # Make sure setup works OK
    FF.DiverseConfGen(rmsd_cutoff, conf_cutoff, energy_cutoff, confab_verbose)
    FF.GetConformers(obmol)
    confs_to_write = obmol.NumConformers()
    obconversion = ob.OBConversion()
    obconversion.SetOutFormat(output_format)
    output_strings = []
    for conf_num in range(confs_to_write):
        obmol.SetConformer(conf_num)
        output_strings.append(obconversion.WriteString(obmol))
    return output_strings


def functionalize(OBmol,functional_group='C',smiles_inds=[0]):
    """functionalize functionalization routine

    Parameters
    ----------
    OBmol : ob.OBMol
        Un"built" 3D ligand
    functional_group : str, optional
        smiles string or name of functional_group, by default 'C'
    smiles_inds : list, optional
        indices where the functional_group should be added, by default [0]

    Returns
    -------
    OBmol : ob.OBMol
        functionalized (unbuilt 3D ligand)
    """

    if functional_group in io_ptable.functional_groups_dict:
        functional_group = io_ptable.functional_groups_dict[functional_group]

    second_mol = get_obmol_smiles(functional_group,build=False,neutralize=False,addHydrogens=False)

    for idx in smiles_inds:
        start_index = OBmol.NumAtoms()
        for i,atom in enumerate(ob.OBMolAtomIter(second_mol)):
            OBmol.AddAtom(atom)
            
        for obbond in ob.OBMolBondIter(second_mol):
            OBmol.AddBond(obbond.GetBeginAtomIdx()+start_index,obbond.GetEndAtomIdx()+start_index,obbond.GetBondOrder())
            
        OBmol.AddBond(start_index+1,idx+1,1)
        for i,atom in enumerate(ob.OBMolAtomIter(OBmol)):
            if (i == start_index) or (i == idx):
                atom.SetImplicitHCount(atom.GetImplicitHCount()-1)

    return OBmol

def convert_obmol_ase(OBMol,posits=None,set_zero=False,add_hydrogens=False):
    """convert_obmol_ase 
    convert obmol to ase

    Parameters
    ----------
    OBMol : OBmol
        Obmol to convert to ASE atoms
    posits : np.ndarray, optional
        positions, by default None
    set_zero : bool, optional
        set positions to origin
    add_hydrogens :bool, optional
        add hydrogens before output, default False

    Returns
    -------
    ase_atoms : ase.Atoms
        converted OBmol
    """
    if hasattr(posits,'__len__'):
        last_atom_index = OBMol.NumAtoms()
        # Set metal to zero
        metal_coords = (posits[last_atom_index-1,0],posits[last_atom_index-1,1],posits[last_atom_index-1,2])
        # set coordinates using OBMol to keep bonding info
        for i, atom in enumerate(ob.OBMolAtomIter(OBMol)):
            atom.SetVector(posits[i, 0]-metal_coords[0], posits[i, 1]-metal_coords[1], posits[i, 2]-metal_coords[2])
    if set_zero:
        for i, atom in enumerate(ob.OBMolAtomIter(OBMol)):
            atom.SetVector(0.0, 0.0, 0.0)
    
    if add_hydrogens:
        OBMol.AddHydrogens()
        mmff94_ok = check_mmff_okay(OBMol)

        # Set up force field
        if mmff94_ok:
            FF = ob.OBForceField.FindForceField("mmff94")
        else:
            FF = ob.OBForceField.FindForceField('UFF')
        FF.Setup(OBMol)

        # Optimize energy
        FF.ConjugateGradients(2000,1e-6)
        FF.GetCoordinates(OBMol)

    # Convert to ASE
    obConversion = ob.OBConversion()
    obConversion.SetOutFormat('xyz')
    xyzStr = obConversion.WriteString(OBMol).strip()
    f = StringIO(xyzStr)
    ase_atoms = read(f,format='xyz',parallel=False)
    return ase_atoms


def convert_ase_obmol(ase_atoms):
    """convert_obmol_ase 
    convert obmol to ase

    Parameters
    ----------
    ase_atoms : ase.Atoms
        structure to convert to openbabel

    Returns
    -------
    OBMol : ob.OBMol
        converted ASE atoms
    """

    OBMol = ob.OBMol()
    for atom in ase_atoms:
        obatom = ob.OBAtom()
        obatom.SetAtomicNum(int(atom.number))
        obatom.SetVector(atom.x,atom.y,atom.z)
        OBMol.AddAtom(obatom)
    return OBMol


def obmol_opt(structure,center_metal=False,fix_m_neighbors=True,
              return_energy=False):
    """obmol_opt take in a structure and optimize with openbabel
    return as ase atoms as default
    Will default to MMFF94 if it is applicable - otherwise it is UFF.

    Parameters
    ----------
    structure : ase.atoms.Atoms
        Structure to optimize with UFF / mmff94
    center_metal : bool, optional
        Move the metal to (0,0,0), default False
    fix_m_neighbors : bool, optional
        Fix the metal neighbors during optimization?, default False
    return_energy : bool, optional
        Return the energy of the optimized structure?, default False

    Returns
    ----------
    out_atoms : ase.atoms.Atoms
        structure
    energy : float, optional
        energy of the structure in UFF / MMFF94 if applicable.
    """
    if isinstance(structure,ase.atoms.Atoms):
        OBMol = convert_ase_obmol(structure)
    elif isinstance(structure,str):
        if 'TRIPOS' in structure:
            OBMol = convert_mol2_obmol(structure,readstring=True)
        elif structure[-5:] == '.mol2':
            OBMol = convert_mol2_obmol(structure,readstring=False)
    elif isinstance(structure,architector.io_molecule.Molecule):
        mol2str = structure.write_mol2('cool.mol2', writestring=True)
        OBMol = convert_mol2_obmol(mol2str, readstring=True)

    mmff94_ok = check_mmff_okay(OBMol)

    # Set up force field
    if mmff94_ok:
        FF = ob.OBForceField.FindForceField("mmff94")
    else:
        FF = ob.OBForceField.FindForceField('UFF')
    
    if fix_m_neighbors:
        _,anums,graph = get_OBMol_coords_anums_graph(OBMol, return_coords=False, get_types=False)
        syms = [io_ptable.elements[x] for x in anums]
        mets = [i for i,x in enumerate(syms) if x in io_ptable.all_metals]
        if len(mets) == 1: # Freeze metal and neighbor positions - relax ligands
            frozen_atoms = [mets[0]+1] + (np.nonzero(np.ravel(graph[mets[0]]))[0] + 1).tolist()
            constr = ob.OBFFConstraints()
            for j in frozen_atoms:
                constr.AddAtomConstraint(int(j))
        elif len(mets) > 1:
            constr = ob.OBFFConstraints()
            print('Warning : Multiple Metals present for FF optimization.')
        elif len(mets) == 0:
            constr = ob.OBFFConstraints()
            # print('No Metals present for FF optimization.')
        FF.Setup(OBMol,constr)
    else:
        FF.Setup(OBMol)

    # Optimize energy
    FF.ConjugateGradients(2000,1e-6)
    FF.GetCoordinates(OBMol)
    energy = FF.Energy()
    if mmff94_ok:
        energy = energy * units.kcal / units.mol
    else:
        energy = energy * units.kJ / units.mol
    out_atoms = convert_obmol_ase(OBMol)
    if center_metal:
        m_ind = [i for i,x in enumerate(out_atoms.get_chemical_symbols()) if x in io_ptable.all_metals]
        if len(m_ind)  == 1:
            new_posits = out_atoms.get_positions()-out_atoms.get_positions()[m_ind]
            out_atoms.set_positions(new_posits)
    if return_energy:
        return out_atoms,energy
    else:
        return out_atoms


def obmol_energy(structure):
    """obmol_energy take in a structure and use openbabel
    to evaluate energy with UFF/MMFF94.

    Parameters
    ----------
    structure : ase.atoms.Atoms
        UFF / mmff94 optimized structure.
    """
    if isinstance(structure,ase.atoms.Atoms):
        OBMol = convert_ase_obmol(structure)
    elif isinstance(structure,str):
        if 'TRIPOS' in structure:
            OBMol = convert_mol2_obmol(structure,readstring=True)
        elif structure[-5:] == '.mol2':
            OBMol = convert_mol2_obmol(structure,readstring=False)
    elif isinstance(structure,architector.io_molecule.Molecule):
        mol2str = structure.write_mol2('cool.mol2', writestring=True)
        OBMol = convert_mol2_obmol(mol2str, readstring=True)

    mmff94_ok = check_mmff_okay(OBMol)

    # Set up force field
    if mmff94_ok:
        FF = ob.OBForceField.FindForceField("mmff94")
    else:
        FF = ob.OBForceField.FindForceField('UFF')

    energy = FF.Energy()
    if mmff94_ok: # Convert to eV
        energy = energy * units.kcal / units.mol
    else:
        energy = energy * units.kJ / units.mol
    # print(FF.GetUnit()) -> Get units for other FFs potentially
    return energy    


def add_dummy_metal(Conf3D,coordList):
    """add_dummy_metal add a dummy metal to a conformer

    Parameters
    ----------
    Conf3D : ob.OBmol
        ligand conformer, likely built in 2D
    coordList : list
        coordination sites of the ligand to the metal
    """
    dummy_metal = ob.OBAtom() # Add the dummy metal to the OBmol
    dummy_metal.SetAtomicNum(26) # Add arbitrary dummy metal - Fe for now - will be removed later
    ttab = ob.OBTypeTable() # Reset types for replication purposes
    ttab.SetFromType('INT')
    ttab.SetToType('INT')
    atypes_old = []
    for i, atom in enumerate(ob.OBMolAtomIter(Conf3D)):
        atypes_old.append(ttab.Translate(atom.GetType()))
    atypes_old.append('Fe')
    Conf3D.AddAtom(dummy_metal)
    for i, atom in enumerate(ob.OBMolAtomIter(Conf3D)):
        atom.SetType(atypes_old[i])
    for i in coordList:
        Conf3D.AddBond(int(i+1), Conf3D.NumAtoms(), 1)


def convert_mol2_obmol(mol2,readstring=True):
    """convert_mol2_obmol
    mol2 to OBMol

    Parameters
    ----------
    mol2 : str
        either filename or mol2 string
    readstring : bool, optional
        read from string or from file, by default True

    Returns
    -------
    obmol : ob.OBMol
        openbabel of the mol2 file
    """
    conv = ob.OBConversion()
    obmol = ob.OBMol()
    conv.SetInFormat('mol2')
    if readstring:
        conv.ReadString(obmol,mol2)
    else:
        conv.ReadFile(obmol,mol2)
    return obmol


def convert_obmol_mol2(OBmol):
    """convert_obmol_mol2
    OBmol to mol2 string

    Parameters
    ----------
    OBmol : ob.OBMol
        OBMol object

    Returns
    -------
    mol2str : str
        mol2 str of the OBmol 
    """
    conv = ob.OBConversion()
    conv.SetOutFormat('mol2')
    mol2str = conv.WriteString(OBmol)
    return mol2str


def convert_xyz_obmol(xyz,readstring=True):
    """convert_xyz_obmol
    xyz to OBMol

    Parameters
    ----------
    xyz : str
        either filename or xyz string
    readstring : bool, optional
        read from string or from file, by default True

    Returns
    -------
    obmol : ob.OBMol
        openbabel of the xyz file
    """
    conv = ob.OBConversion()
    obmol = ob.OBMol()
    conv.SetInFormat('xyz')
    if readstring:
        conv.ReadString(obmol,xyz)
    else:
        conv.ReadFile(obmol,xyz)
    return obmol


def remove_obmol_metals(Conf3D):
    """remove_obmol_metals remove metals from obmol

    Parameters
    ----------
    Conf3D : ob.OBMol
        openbabel molecule with a metal atom

    Returns
    -------
    Conf3D : ob.OBMol
        moleucle with no metal atoms remaining
    """
    rem_list = []
    for i, atom in enumerate(ob.OBMolAtomIter(Conf3D)):
        anum = atom.GetAtomicNum()
        if io_ptable.elements[anum] in io_ptable.all_metals:
            rem_list.append(i+1)
    rem_list = sorted(rem_list)[::-1]
    for ind in rem_list:
        metal_atom = Conf3D.GetAtom(ind)
        Conf3D.DeleteAtom(metal_atom)
    return Conf3D


def obmol_lig_split(mol2string,return_info=False,calc_coord_atoms=True):
    """obmol_lig_split 
    Take in a mol2string and use openbabel to split into ligands, convert to smiles, 
    and calculate metal-ligand coordinating atoms implicit in the mol2string.

    Parameters
    ----------
    mol2string : str
       mol2string of ideally mononuclear metal complex.
    return_info : bool, optional
        return information dictionary as well with metal,ligand_charges,coordAt symbols, default False
    calc_coord_atoms : bool, optional
        return the coordination atoms for smiles, default True

    Returns
    -------
    ligand_smiles : list(str)
        list of the ligand smiles present in the complex
    coord_atom_lists : list(list)
        list of coordinating atom indices for the smiles str
    """
    obmol = convert_mol2_obmol(mol2string)
    _,anums,graph = get_OBMol_coords_anums_graph(obmol)
    bo_dict, _ = get_OBMol_bo_dict_atom_types(obmol,metal_passed=False)
    met_inds = [i for i,x in enumerate(anums) if (io_ptable.elements[x] in io_ptable.all_metals)]
    shape = graph.shape
    only_mets_graph = np.zeros(shape)
    init_graph = graph.copy()
    # Create 2 graphs - one with metals removed, one with only the metals graphs
    for ind in sorted(met_inds):
        only_mets_graph[:,ind] = graph[ind]
        only_mets_graph[ind,:] = graph[ind]
        # Zero out the deleted inds
        init_graph[ind,:] = np.zeros(shape[0])
        init_graph[:,ind] = np.zeros(shape[0])
    csg = csgraph.csgraph_from_dense(init_graph)
    # Break apart zeroed graph into connected components
    disjoint_components = csgraph.connected_components(csg)[1]
    ligs_inds = []
    for ind in sorted(list(set(disjoint_components))): # sort for reproducability
        subgraph = np.where(disjoint_components == ind)[0]
        sg = np.array([x for x in subgraph if x not in met_inds]) # Check not deleted atoms
        sg.sort()
        if len(sg) > 0:
            ligs_inds.append(sg)
    ligand_smiles = []
    coord_atom_lists = []
    for lig in ligs_inds:
        lig = lig.tolist()
        ligobmol = ob.OBMol()
        coord_atom_list = []
        for i,atom_ind in enumerate(lig):
            atom_ind += 1
            for l, atom in enumerate(ob.OBMolAtomIter(obmol)):
                if (l+1 == atom_ind): 
                    ligobmol.AddAtom(atom)
            for k in bo_dict.keys():
                if (atom_ind in k):
                    other_ind = [x for x in k if x!=atom_ind][0] - 1
                    if other_ind in met_inds:
                        coord_atom_list.append(i)
        for k in bo_dict.keys():
            if (k[0]-1 in lig) and (k[1]-1 in lig):
                start_ind = lig.index(k[0]-1) + 1
                end_ind = lig.index(k[1]-1) + 1
                ligobmol.AddBond(start_ind,end_ind,bo_dict[k])
        ligobmol.PerceiveBondOrders()
        # Key block for catching where coordinating atoms were deprotonated
        ### WORKING -> Does not work great for nitrogen compounds.
        for l, atom in enumerate(ob.OBMolAtomIter(ligobmol)):
            total_val = (io_ptable.valence_electrons[atom.GetAtomicNum()] + atom.GetTotalValence())
            close = np.argmin(np.abs(np.array(io_ptable.filled_valence_electrons)-total_val))
            atom.SetFormalCharge(int(atom.GetFormalCharge()-(io_ptable.filled_valence_electrons[close]-total_val)))
        new_smiles = get_smiles_obmol(ligobmol,canonicalize=True)
        ligand_smiles.append(new_smiles)
        if calc_coord_atoms:
            new_coord_atom_list = map_coord_ats_smiles(new_smiles, ligobmol, coord_atom_list)
            coord_atom_lists.append(sorted(new_coord_atom_list))
        else:
            coord_atom_lists = []
    if not return_info:
        return ligand_smiles,coord_atom_lists
    else:
        info_dict = dict()
        if len(met_inds) > 0:
            info_dict['metal'] = io_ptable.elements[anums[met_inds[0]]]
            info_dict['metal_ind'] = met_inds[0]
        else:
            info_dict['metal'] = None
            info_dict['metal_ind'] = None
        lig_obmols = [get_obmol_smiles(smi,addHydrogens=True,
                    neutralize=False,
                    build=False) for smi in ligand_smiles]
        info_dict['lig_charges'] = [x.GetTotalCharge() for x in lig_obmols]
        lig_coord_ats = []
        if calc_coord_atoms:
            for i,lig_obmol in enumerate(lig_obmols):
                _,anums,_ = get_OBMol_coords_anums_graph(lig_obmol,get_types=False)
                lig_coord_ats.append(','.join([io_ptable.elements[x] for x in np.array(anums)[np.array(coord_atom_lists[i])]]))
            info_dict['lig_coord_ats'] = lig_coord_ats
        else:
            info_dict['lig_coord_ats'] = None
        return ligand_smiles,coord_atom_lists, info_dict


def get_canonical_label(obmol):
    """Create a canonical label for the molecule using pynauty.
    Warning - can fail/hang on particularly gnarly graphs.
    """
    _,anums,graph = get_OBMol_coords_anums_graph(obmol,get_types=True)
    grph = pnGraph(graph.shape[0])
    colors = get_vertex_coloring(anums) # Color by atom_type
    for n in range(graph.shape[0]):
        neighs = np.nonzero(np.ravel(graph[n]))[0]
        grph.connect_vertex(n, neighs.tolist())
    grph.set_vertex_coloring(colors)
    can_label = canon_label(grph) # Pynauty canonical label
    return can_label


def get_vertex_coloring(anums):
    """ Create set of like atoms by atomic number for graph labelling
    """
    if isinstance(anums,list): # Convert to array
        anums = np.array(anums)
    sym_labels = sorted(list(set(anums)))
    out_sets = []
    for anum in sym_labels:
        ind_set = set(np.where(anums == anum)[0])
        out_sets.append(ind_set)
    return out_sets


def map_coord_ats_smiles(lig_smiles, lig_obmol, coord_atoms):
    """map_coord_ats_smiles 
    Map the 3d structure of a ligand with encoded coordinating atom information
    to the smiles string of the ligand to get the correct smicat atoms
    Basic routine is done by converting both smiles and 3d structure to 
    molecular graphs, then using the coordinating atom information in addition 
    to the pynauty canonicalized form of the molecular graphs (colored by atom type)
    to map the 3d structure graph to the smiles graph.
    """
    tmp1 = get_obmol_smiles(lig_smiles)
    can1 = get_canonical_label(tmp1) 
    can2 = get_canonical_label(lig_obmol) # Know indices
    smicat = []
    for atom in coord_atoms: # Match canonicalized graph indices
        a_ind = can2.index(atom)
        smicat.append(can1[a_ind])
    return smicat

def get_fingerprint(obmol,fp='FP2'):
    """get_fingerprint 
    Gets the fingerprint for an obmol molecule.

    Parameters
    ----------
    obmol : ob.OBMol
        molecule
    fp : str, optional
        type of fingerprint, by default 'FP2'

    Returns
    -------
    fp : bit vector
        openbabel fingerprint of the molecule
    """
    pybelmol = pybel.Molecule(obmol)
    fp = pybelmol.calcfp(fp)
    return fp

# Main
if (__name__ == '__main__'):
    # Variables
    smiles = "NCCN"
    print(smiles2xyz(smiles))